package utility;

import java.io.File;
import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import org.semanticweb.owl.apibinding.OWLManager;
import org.semanticweb.owl.model.OWLClass;
import org.semanticweb.owl.model.OWLDataProperty;
import org.semanticweb.owl.model.OWLObjectProperty;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.model.OWLOntologyCreationException;
import org.semanticweb.owl.model.OWLOntologyManager;
import complexMapping.ComplexMappingException;
import correspondenceComputations.ComputationSetting;
import de.unima.ki.mmatch.MMatchException;
import de.unima.ki.mmatch.descriptions.Term;

/**
 * 
 * @author Dominique Ritze
 * 
 * The class Ontology contains the classes, properties and some ontology characteristics like the filepath.  
 *
 */
public class Ontology {
	
	private boolean[] delimiter = new boolean[3];
	private Set<OWLClass> classes = new HashSet<OWLClass>();
	private Set<OWLObjectProperty> objectProperties = new HashSet<OWLObjectProperty>();
	private Set<OWLDataProperty> datatypeProperties = new HashSet<OWLDataProperty>();
	private OWLOntology ontology;
	private URI filepath;
	private HashMap<OWLClass, Term> termClasses = new HashMap<OWLClass, Term>();
	private HashMap<OWLObjectProperty, Term> termObjectProperties = new HashMap<OWLObjectProperty, Term>();
	private HashMap<OWLDataProperty, Term> termDatatypeProperties = new HashMap<OWLDataProperty, Term>();
	
	/**
	 * Create an ontology out of the filepath and build all important structures (like set of classes or properties).
	 * 
	 * @param filepath
	 * @throws ComplexMappingException 
	 */
	public Ontology(String filepath) throws ComplexMappingException {
		
		this.ontology = buildOntology(filepath);		
		this.classes = ontology.getReferencedClasses();
		this.objectProperties = ontology.getReferencedObjectProperties();
		this.datatypeProperties = ontology.getReferencedDataProperties();
		this.delimiter = getDelimiter(ontology);
		File file = new File(filepath);
		this.filepath = file.toURI();
		
		try {
			//build a hash map with OWLClasses and the corresponding terms
			for(OWLClass cls : classes) {
				if(cls.getURI().getFragment()!= null) {
					termClasses.put(cls, new Term(cls.getURI().getFragment(), delimiter));
				}
				//avoid some problems with URIs which does not contain a #
				else {
					String[] realName = cls.getURI().toString().split("/");
					termClasses.put(cls, new Term(realName[realName.length-1], delimiter));
				}
			}
			//build the object property - term map
			for(OWLObjectProperty prop : objectProperties) {
				if(prop.getURI().getFragment() != null) {
					termObjectProperties.put(prop, new Term(prop.getURI().getFragment(), delimiter, true));
				}
				else {
					String[] realName = prop.getURI().toString().split("/");
					termObjectProperties.put(prop, new Term(realName[realName.length-1], delimiter, true));
				}
			}
			//build the datatype property - term map
			for(OWLDataProperty prop : datatypeProperties) {
				if(prop.getURI().getFragment() !=null) {
					termDatatypeProperties.put(prop, new Term(prop.getURI().getFragment(), delimiter, true));
				}	
				else {
					String[] realName = prop.getURI().toString().split("/");
					termDatatypeProperties.put(prop, new Term(realName[realName.length-1], delimiter, true));
				}
			}
		} catch(MMatchException e) {
			throw new ComplexMappingException(ComplexMappingException.MMATCH_EXCEPTION, "Could not create maps with terms in ontology.", e);
		}
		
	}
	
	/**
	 * Compute the delimiter which occur in the ontology.
	 * 
	 * @param ontology
	 * @return Array with the real delimiter, first entry: hyphen, second: blank, third: underscore. The entry is true if the
	 * character should be a real delimiter.
	 */
	public static boolean[] getDelimiter(OWLOntology ontology) {
		int countHyphen = 0;
		int countUnderscore = 0;
		int countBlank = 0;	
		int size = ontology.getReferencedClasses().size();
		
		/*
		 * The following three loops check which delimiters are used in the ontology. 
		 * If occurence delimiter devided by number of all delimiters is less than 0.05
		 * the delimiter is not considered as a real delimiter.
		 */
		//iterate through all classes of the ontology and count the hypens, the blanks and the underscores
		for(OWLClass cls : ontology.getReferencedClasses()) {
			if(cls.toString().contains("-")) {
				countHyphen++;
			}
			if(cls.toString().contains(" ")) {
				countBlank++;
			}
			if(cls.toString().contains("_")) {
				countUnderscore++;
			}
		}
        	
		//iterate through all data properties and count hypens, blanks, underscores
		for(OWLDataProperty dataProp : ontology.getReferencedDataProperties()) {
			if(dataProp.toString().contains("-")) {
				countHyphen++;
			}
			if(dataProp.toString().contains(" ")) {
				countBlank++;
			}
			if(dataProp.toString().contains("_")) {
				countUnderscore++;
			}
		}
        	
		//iterate through all object properties and count hypens, blanks, underscores
		for(OWLObjectProperty objectProp : ontology.getReferencedObjectProperties()) {  
			if(objectProp.toString().contains("-")) {
				countHyphen++;
			}
			if(objectProp.toString().contains(" ")) {
				countBlank++;
			}
			if(objectProp.toString().contains("_")) {
				countUnderscore++;
			} 
		}
		boolean[] delimiter = new boolean[]{false, false, false};
		if((double)countHyphen/(double)size > ComputationSetting.getDelimiterPercent()) {
			delimiter[0] = true;
		}
		if((double)countBlank/(double)size > ComputationSetting.getDelimiterPercent()) {
			delimiter[1] = true;
		}
		if((double)countUnderscore/(double)size > ComputationSetting.getDelimiterPercent()) {
			delimiter[2] = true;
		}
		return delimiter;
	}
	
	/**
	 * Build an OWLOntology out of a filepath.
	 * 
	 * @param filepath
	 * @return
	 * @throws ComplexMappingException 
	 */
	public static OWLOntology buildOntology(String filepath) throws ComplexMappingException {
		//create the OWLOntologies out of the filepaths
		OWLOntologyManager manager1 = OWLManager.createOWLOntologyManager();    	
    	File ontologyFile = new File(filepath);        
    	try {
    		return manager1.loadOntologyFromPhysicalURI(URI.create(ontologyFile.toURI().toString())); 
    	} catch(OWLOntologyCreationException e) {
    		throw new ComplexMappingException(ComplexMappingException.CREATION_EXCEPTION, "Could not create an OWLOntology in Ontology.", e);
    	}
    	
    	   	
	}

	public boolean[] getDelimiter() {
		return delimiter;
	}

	public Set<OWLClass> getClasses() {
		return classes;
	}

	public Set<OWLObjectProperty> getObjectProperties() {
		return objectProperties;
	}

	public Set<OWLDataProperty> getDatatypeProperties() {
		return datatypeProperties;
	}

	public OWLOntology getOntology() {
		return ontology;
	}

	public URI getFilepath() {
		return filepath;
	}

	public HashMap<OWLClass, Term> getTermClasses() {
		return termClasses;
	}

	public HashMap<OWLObjectProperty, Term> getTermObjectProperties() {
		return termObjectProperties;
	}

	public HashMap<OWLDataProperty, Term> getTermDatatypeProperties() {
		return termDatatypeProperties;
	}
}
