package com.clarkparsia.modularity.locality;

import org.semanticweb.owl.model.*;

import java.util.Collection;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.Set;


/**
 * <p/>
 * Title: </p>
 * <p/>
 * Description: Implements syntactic locality evaluation for axioms </p>
 * <p/>
 * Copyright: Copyright (c) 2007 </p>
 * <p/>
 * Company: Clark & Parsia, LLC. <http://www.clarkparsia.com> </p>
 *
 * @author Mike Smith
 */
public class SyntacticLocalityEvaluator implements LocalityEvaluator {

    private LocalityClass localityCls;

    private AxiomLocalityVisitor axiomVisitor;

    private static EnumSet<LocalityClass> supportedLocalityClasses = EnumSet.of(LocalityClass.TOP_BOTTOM,
                                                                                LocalityClass.BOTTOM_BOTTOM,
                                                                                LocalityClass.TOP_TOP);


    public SyntacticLocalityEvaluator(LocalityClass localityClass) {
        this.localityCls = localityClass;
        this.axiomVisitor = new AxiomLocalityVisitor();
        if (!supportedLocalityClasses.contains(localityClass)) {
            throw new RuntimeException("Unsupported locality class: " + localityClass);
        }
    }


    public Set<LocalityClass> supportedLocalityClasses() {
        return supportedLocalityClasses;
    }


    private class AxiomLocalityVisitor implements OWLAxiomVisitor {

        private BottomEquivalenceEvaluator bottomEvaluator;

        private boolean isLocal;

        private Collection<? extends OWLEntity> signature;

        private TopEquivalenceEvaluator topEvaluator;


        public AxiomLocalityVisitor() {
            bottomEvaluator = new BottomEquivalenceEvaluator();
            topEvaluator = new TopEquivalenceEvaluator();
            topEvaluator.setBottomEvaluator(bottomEvaluator);
            bottomEvaluator.setTopEvaluator(topEvaluator);
        }


        public boolean isLocal(OWLAxiom axiom, Collection<? extends OWLEntity> signature) {
            this.signature = signature;
            isLocal = false;
            axiom.accept(this);
            return isLocal;
        }


        // TODO: (TS) Is this correct for the *_BOTTOM case? The empty object property is antisymmetric!
        public void visit(OWLAntiSymmetricObjectPropertyAxiom axiom) {
            isLocal = false;
        }


        //TODO: (TS) Shouldn't locality of the annotation axiom depend on locality of the annotated axiom?
        public void visit(OWLAxiomAnnotationAxiom axiom) {
            isLocal = true;
        }


        //TODO: (TS) Any individual declaration is non-local, but a class assertion axiom can be local?
        public void visit(OWLClassAssertionAxiom axiom) {
            isLocal = topEvaluator.isTopEquivalent(axiom.getDescription(), signature, localityCls);
        }


        //TODO: (TS) Any individual declaration is non-local, but a dp assertion axiom can be local?
        public void visit(OWLDataPropertyAssertionAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = false;
                    break;
                case TOP_TOP:
                    isLocal = !signature.contains(axiom.getProperty().asOWLDataProperty());
                    break;
            }
        }


        public void visit(OWLDataPropertyDomainAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getProperty().asOWLDataProperty()) || topEvaluator.isTopEquivalent(
                            axiom.getDomain(),
                            signature,
                            localityCls);
                    break;
                case TOP_TOP:
                    isLocal = topEvaluator.isTopEquivalent(axiom.getDomain(), signature, localityCls);
                    break;
            }
        }


        // TODO: (TS) Why is this treated differently from the previous case?
        public void visit(OWLDataPropertyRangeAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isLocal = false;
                    break;
            }
        }


        public void visit(OWLDataSubPropertyAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getSubProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isLocal = !signature.contains(axiom.getSuperProperty().asOWLDataProperty());
                    break;
            }
        }


        public void visit(OWLDeclarationAxiom axiom) {
            isLocal = !(axiom.getEntity().isOWLIndividual());
        }


        public void visit(OWLDifferentIndividualsAxiom axiom) {
            isLocal = false;
        }


        //TODO: (TS) An n-ary disj classes axioms is now non-local iff at least two of the involved class descriptions are not bot-equivalent.
        public void visit(OWLDisjointClassesAxiom axiom) {
            Collection<OWLDescription> disjs = axiom.getDescriptions();
            int size = disjs.size();
            if (size == 1) {
                throw new RuntimeException("Unary disjoint axiom.");
            }
            else {
                boolean nonBottomEquivDescFound = false;
                for (OWLDescription desc : disjs) {
                    if (!bottomEvaluator.isBottomEquivalent(desc, signature, localityCls)) {
                        if (nonBottomEquivDescFound) {
                            isLocal = false;
                            return;
                        }
                        else {
                            nonBottomEquivDescFound = true;
                        }
                    }
                }
            }
            isLocal = true;
        }

        //TODO: (TS) Continue asking Uli+Bijan
        //TODO: (TS) In the *_BOTTOM case, shouldn't this be the same case distinction as for disj classes axioms?
        public void visit(OWLDisjointDataPropertiesAxiom axiom) {
            isLocal = false;
        }


        //TODO: (TS) In the *_BOTTOM case, shouldn't this be the same case distinction as for disj classes axioms?
        public void visit(OWLDisjointObjectPropertiesAxiom axiom) {
            isLocal = false;
        }


        //TODO: (TS) There are cases where a disj union axiom is local, e.g., when all involved class descriptions are bot-equivalent.
        public void visit(OWLDisjointUnionAxiom axiom) {
            isLocal = false;
        }


        //TODO: (TS) Shouldn't locality of the annotation axiom depend on the fact whether the annotated entity occurs in some non-local axiom?
        public void visit(OWLEntityAnnotationAxiom axiom) {
            isLocal = true;
        }


        //TODO: (TS) Can we make this more efficient?
        public void visit(OWLEquivalentClassesAxiom axiom) {
            isLocal = true;

            Iterator<OWLDescription> eqs = axiom.getDescriptions().iterator();
            OWLDescription first = eqs.next();

            // axiom is local if it contains a single class description
            if (!eqs.hasNext())
                return;

            // axiom is local iff either all concepts evaluate to TOP or all
            // evaluate to BOTTOM

            // check if first concept is BOTTOM
            boolean isBottom = bottomEvaluator.isBottomEquivalent(first, signature, localityCls);

            // if not BOTTOM or not TOP then this axiom is non-local
            if (!isBottom && !topEvaluator.isTopEquivalent(first, signature, localityCls))
                isLocal = false;

            // unless we find a non-locality process all the class descriptions
            while (isLocal && eqs.hasNext()) {
                OWLDescription next = eqs.next();

                if (isBottom) {
                    // other concepts were BOTTOM so this one should be BOTTOM
                    // too
                    if (!bottomEvaluator.isBottomEquivalent(next, signature, localityCls)) {
                        isLocal = false;
                    }
                }
                else {
                    // other concepts were TOP so this one should be TOP too
                    if (!topEvaluator.isTopEquivalent(next, signature, localityCls)) {
                        isLocal = false;
                    }
                }
            }
        }


        public void visit(OWLEquivalentDataPropertiesAxiom axiom) {
            Collection<OWLDataPropertyExpression> eqs = axiom.getProperties();
            int size = eqs.size();
            if (size == 1) {
                isLocal = true;
            }
            else {
                for (OWLDataPropertyExpression p : eqs) {
                    if (signature.contains(p.asOWLDataProperty())) {
                        isLocal = false;
                        return;
                    }
                }
                isLocal = true;
            }
        }


        public void visit(OWLEquivalentObjectPropertiesAxiom axiom) {
            Collection<OWLObjectPropertyExpression> eqs = axiom.getProperties();
            int size = eqs.size();
            if (size == 1) {
                isLocal = true;
            }
            else {
                for (OWLObjectPropertyExpression p : eqs) {
                    if (signature.contains(p.getNamedProperty())) {
                        isLocal = false;
                        return;
                    }
                }
                isLocal = true;
            }
        }


        public void visit(OWLFunctionalDataPropertyAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isLocal = false;
                    break;
            }
        }


        public void visit(OWLFunctionalObjectPropertyAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getProperty().asOWLObjectProperty());
                    break;
                case TOP_TOP:
                    isLocal = false;
                    break;
            }
        }


        public void visit(OWLImportsDeclaration axiom) {
            isLocal = false;
        }


        public void visit(OWLInverseFunctionalObjectPropertyAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getProperty().getNamedProperty());
                    break;
                case TOP_TOP:
                    isLocal = false;
                    break;
            }
        }


        public void visit(OWLInverseObjectPropertiesAxiom axiom) {
            isLocal = !signature.contains(axiom.getFirstProperty().getNamedProperty()) && !signature.contains(axiom.getSecondProperty().getNamedProperty());
        }


        // TODO: (TS) Is this correct for the *_BOTTOM case? The empty object property is irreflexive!
        public void visit(OWLIrreflexiveObjectPropertyAxiom axiom) {
            isLocal = false;
        }


        // TODO: (TS) Why isn't this treated dually to the positive case? I.e., the axiom should be bot-local if the property is not in the signature.
        public void visit(OWLNegativeDataPropertyAssertionAxiom axiom) {
            isLocal = false;
        }


        // TODO: (TS) Why isn't this treated dually to the positive case? I.e., the axiom should be bot-local if the property is not in the signature. 
        public void visit(OWLNegativeObjectPropertyAssertionAxiom axiom) {
            isLocal = false;
        }


        public void visit(OWLObjectPropertyAssertionAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = false;
                    break;
                case TOP_TOP:
                    isLocal = !signature.contains(axiom.getProperty().getNamedProperty());
                    break;
            }
        }


        //TODO: (TS) Shouldn't this be local iff (a) at least one of the OPs in the chain is bot-equiv or (b) all involved OPs are top-equiv?
        public void visit(OWLObjectPropertyChainSubPropertyAxiom axiom) {
            // TODO: Can be local in TOP_TOP case -- (TS) Also in the *_BOTTOM case!
            isLocal = false;
        }


        public void visit(OWLObjectPropertyDomainAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getProperty().getNamedProperty()) || topEvaluator.isTopEquivalent(
                            axiom.getDomain(),
                            signature,
                            localityCls);
                    break;
                case TOP_TOP:
                    isLocal = topEvaluator.isTopEquivalent(axiom.getDomain(), signature, localityCls);
                    break;
            }
        }


        public void visit(OWLObjectPropertyRangeAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getProperty().getNamedProperty()) || topEvaluator.isTopEquivalent(
                            axiom.getRange(),
                            signature,
                            localityCls);
                    break;
                case TOP_TOP:
                    isLocal = topEvaluator.isTopEquivalent(axiom.getRange(), signature, localityCls);
                    break;
            }
        }


        public void visit(OWLObjectSubPropertyAxiom axiom) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isLocal = !signature.contains(axiom.getSubProperty().getNamedProperty());
                    break;
                case TOP_TOP:
                    isLocal = !signature.contains(axiom.getSuperProperty().getNamedProperty());
                    break;
            }
        }


        public void visit(OWLOntologyAnnotationAxiom axiom) {
            isLocal = true;
        }


        public void visit(OWLReflexiveObjectPropertyAxiom axiom) {
            isLocal = !signature.contains(axiom.getProperty().getNamedProperty());
        }


        public void visit(OWLSameIndividualsAxiom axiom) {
            isLocal = false;
        }


        public void visit(OWLSubClassAxiom axiom) {
            isLocal = bottomEvaluator.isBottomEquivalent(axiom.getSubClass(),
                                                         signature,
                                                         localityCls) || topEvaluator.isTopEquivalent(axiom.getSuperClass(),
                                                                                                      signature,
                                                                                                      localityCls);
        }


        public void visit(OWLSymmetricObjectPropertyAxiom axiom) {
            isLocal = !signature.contains(axiom.getProperty().getNamedProperty());
        }


        public void visit(OWLTransitiveObjectPropertyAxiom axiom) {
            isLocal = !signature.contains(axiom.getProperty().getNamedProperty());
        }


        //TODO: (TS) Can't we treat this in a more differentiated way?
        public void visit(SWRLRule axiom) {
            isLocal = false;
        }
    }


    /**
     * Used to determine if descriptions are equivalent to \bottom using the provided locality class
     */
    private static class BottomEquivalenceEvaluator implements OWLDescriptionVisitor {

        private boolean isBottomEquivalent;

        private LocalityClass localityCls;

        private Collection<? extends OWLEntity> signature;

        private TopEquivalenceEvaluator topEvaluator;


        public BottomEquivalenceEvaluator() {
        }


        private boolean isBottomEquivalent(OWLDescription desc) {
            desc.accept(this);
            return isBottomEquivalent;
        }


        public boolean isBottomEquivalent(OWLDescription desc, Collection<? extends OWLEntity> signature,
                                          LocalityClass localityCls) {
            this.localityCls = localityCls;
            this.signature = signature;
            desc.accept(this);
            return isBottomEquivalent;
        }


        public void setTopEvaluator(TopEquivalenceEvaluator evaluator) {
            this.topEvaluator = evaluator;
        }


        public void visit(OWLClass desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                    isBottomEquivalent = desc.isOWLNothing() || (!desc.isOWLThing() && !signature.contains(desc));
                    break;
                case TOP_BOTTOM:
                case TOP_TOP:
                    isBottomEquivalent = desc.isOWLNothing();
                    break;
            }
        }


        public void visit(OWLDataAllRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = false;
                    break;
                case TOP_TOP:
                    // FIXME: This ignores TOP_DATA case
                    // TODO: (TS) This only works if, in addition, desc.getFiller is not (equivalent to?) rdfs:Literal. (?)
                    isBottomEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
            }
        }


        // TODO: (TS) Doesn't the answer in the TOP_TOP case depend on the cardinality of the data range? I'd say it's never bottom equivalent.
        // (Check this for Max and Min restrictions, too.)
        public void visit(OWLDataExactCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    // TODO: Special handling for the n == 0 case
                    isBottomEquivalent = false;
                    break;
                case TOP_TOP:
                    isBottomEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
            }
        }


        public void visit(OWLDataMaxCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    // TODO: (TS) A max restriction is never bottom equivalent in this case.
                    isBottomEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty()) && (desc.getCardinality() == 0);
                    break;
            }
        }


        public void visit(OWLDataMinCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    // TODO: Special handling for the n == 0 case
                    isBottomEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = false;
                    break;
            }
        }


        public void visit(OWLDataSomeRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = false;
                    break;
            }
        }


        public void visit(OWLDataValueRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = false;
                    break;
            }
        }


        public void visit(OWLObjectAllRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = false;
                    break;
                case TOP_TOP:
                    isBottomEquivalent = isBottomEquivalent(desc.getFiller());
                    break;
            }
        }


        public void visit(OWLObjectComplementOf desc) {
            isBottomEquivalent = topEvaluator.isTopEquivalent(desc.getOperand(), signature, localityCls);
        }


        // TODO: (TS) Check ALL object/data cardinality restriction cases in both evaluators!!!
        public void visit(OWLObjectExactCardinalityRestriction desc) {
            // TODO: Special handling for the n == 0 case
            isBottomEquivalent = false;
        }


        public void visit(OWLObjectIntersectionOf desc) {
            for (OWLDescription conj : desc.getOperands()) {
                if (isBottomEquivalent(conj)) {
                    isBottomEquivalent = true;
                    return;
                }
            }
            isBottomEquivalent = false;
        }


        public void visit(OWLObjectMaxCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = !signature.contains(desc.getProperty().getNamedProperty()) || isBottomEquivalent(
                            desc.getFiller());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = !signature.contains(desc.getProperty().getNamedProperty()) && (desc.getCardinality() == 0);
                    break;
            }
        }


        public void visit(OWLObjectMinCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    // TODO: Special handling for the n == 0 case
                    isBottomEquivalent = !signature.contains(desc.getProperty().getNamedProperty());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = false;
                    break;
            }
        }


        public void visit(OWLObjectOneOf desc) {
            isBottomEquivalent = desc.getIndividuals().isEmpty();
        }


        public void visit(OWLObjectSelfRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = !signature.contains(desc.getProperty().getNamedProperty());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = false;
                    break;
            }
        }


        public void visit(OWLObjectSomeRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = !signature.contains(desc.getProperty().getNamedProperty()) || isBottomEquivalent(
                            desc.getFiller());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = isBottomEquivalent(desc.getFiller());
                    break;
            }
        }


        public void visit(OWLObjectUnionOf desc) {
            for (OWLDescription disj : desc.getOperands()) {
                if (!isBottomEquivalent(disj)) {
                    isBottomEquivalent = false;
                    return;
                }
            }
            isBottomEquivalent = true;
        }


        public void visit(OWLObjectValueRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isBottomEquivalent = !signature.contains(desc.getProperty().getNamedProperty()) || !signature.contains(
                            desc.getValue());
                    break;
                case TOP_TOP:
                    isBottomEquivalent = false;
                    break;
            }
        }
    }


    /**
     * Used to determine if descriptions are equivalent to \top using the provided locality class
     */
    private static class TopEquivalenceEvaluator implements OWLDescriptionVisitor {

        private BottomEquivalenceEvaluator bottomEvaluator;

        private boolean isTopEquivalent;

        private LocalityClass localityCls;

        private Collection<? extends OWLEntity> signature;


        public TopEquivalenceEvaluator() {
        }


        private boolean isTopEquivalent(OWLDescription desc) {
            desc.accept(this);
            return isTopEquivalent;
        }


        public boolean isTopEquivalent(OWLDescription desc, Collection<? extends OWLEntity> signature,
                                       LocalityClass localityCls) {
            this.localityCls = localityCls;
            this.signature = signature;
            desc.accept(this);
            return isTopEquivalent;
        }


        public void setBottomEvaluator(BottomEquivalenceEvaluator evaluator) {
            this.bottomEvaluator = evaluator;
        }


        public void visit(OWLClass desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                    isTopEquivalent = desc.isOWLThing();
                    break;
                case TOP_BOTTOM:
                case TOP_TOP:
                    isTopEquivalent = desc.isOWLThing() || (!desc.isOWLNothing() && !signature.contains(desc));
                    break;
            }
        }


        public void visit(OWLDataAllRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    // FIXME: This ignores TOP_DATA case
                    isTopEquivalent = false;
                    break;
            }
        }


        public void visit(OWLDataExactCardinalityRestriction desc) {
            // TODO: Special handling for the n == 0 case
            isTopEquivalent = false;
        }


        public void visit(OWLDataMaxCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    // TODO: Special handling for the n == 0 case
                    isTopEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
                case TOP_TOP:
                    isTopEquivalent = false;
                    break;
            }
        }


        public void visit(OWLDataMinCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = false;
                    break;
                case TOP_TOP:
                    isTopEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty()) && (desc.getCardinality() == 0);
                    break;
            }
        }


        public void visit(OWLDataSomeRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = false;
                    break;
                case TOP_TOP:
                    isTopEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
            }
        }


        public void visit(OWLDataValueRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = false;
                    break;
                case TOP_TOP:
                    isTopEquivalent = !signature.contains(desc.getProperty().asOWLDataProperty());
                    break;
            }
        }


        public void visit(OWLObjectAllRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = !signature.contains(desc.getProperty().getNamedProperty()) || isTopEquivalent(desc.getFiller());
                    break;
                case TOP_TOP:
                    isTopEquivalent = isTopEquivalent(desc.getFiller());
                    break;
            }
        }


        public void visit(OWLObjectComplementOf desc) {
            isTopEquivalent = bottomEvaluator.isBottomEquivalent(desc.getOperand(), signature, localityCls);
        }


        public void visit(OWLObjectExactCardinalityRestriction desc) {
            // TODO: Special handling for the n == 0 case
            isTopEquivalent = false;
        }


        public void visit(OWLObjectIntersectionOf desc) {
            for (OWLDescription conj : desc.getOperands()) {
                if (!isTopEquivalent(conj)) {
                    isTopEquivalent = false;
                    return;
                }
            }
            isTopEquivalent = true;
        }


        public void visit(OWLObjectMaxCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    // TODO: Special handling for the n == 0 case
                    isTopEquivalent = !signature.contains(desc.getProperty().getNamedProperty());
                    break;
                case TOP_TOP:
                    isTopEquivalent = false;
                    break;
            }
        }


        public void visit(OWLObjectMinCardinalityRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    // TODO: Special handling for the n == 0 case
                    isTopEquivalent = false;
                    break;
                case TOP_TOP:
                    isTopEquivalent = !signature.contains(desc.getProperty().getNamedProperty()) && (desc.getCardinality() <= 1);
                    break;
            }
        }


        public void visit(OWLObjectOneOf desc) {
            isTopEquivalent = false;
        }


        public void visit(OWLObjectSelfRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = false;
                    break;
                case TOP_TOP:
                    isTopEquivalent = !signature.contains(desc.getProperty().getNamedProperty());
                    break;
            }
        }


        public void visit(OWLObjectSomeRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = false;
                    break;
                case TOP_TOP:
                    isTopEquivalent = !signature.contains(desc.getProperty()) && isTopEquivalent(desc.getFiller());
                    break;
            }
        }


        public void visit(OWLObjectUnionOf desc) {
            for (OWLDescription conj : desc.getOperands()) {
                if (isTopEquivalent(conj)) {
                    isTopEquivalent = true;
                    return;
                }
            }
            isTopEquivalent = false;
        }


        public void visit(OWLObjectValueRestriction desc) {
            switch (localityCls) {
                case BOTTOM_BOTTOM:
                case TOP_BOTTOM:
                    isTopEquivalent = false;
                    break;
                case TOP_TOP:
                    isTopEquivalent = !signature.contains(desc.getProperty().getNamedProperty());
                    break;
            }
        }
    }


    /**
     * True if the axiom is syntactically local w.r.t. given signature
     */
    public boolean isLocal(OWLAxiom axiom, Set<? extends OWLEntity> signature) {
        return axiomVisitor.isLocal(axiom, signature);
    }
}
