package org.semanticweb.owl.profiles;

import org.semanticweb.owl.model.*;
import org.semanticweb.owl.util.OWLObjectPropertyManager;

import java.net.URI;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
/*
 * Copyright (C) 2008, University of Manchester
 *
 * Modifications to the initial code base are copyright of their
 * respective authors, or their employers as appropriate.  Authorship
 * of the modifications may be determined from the ChangeLog placed at
 * the end of this file.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


/**
 * Author: Matthew Horridge<br>
 * The University Of Manchester<br>
 * Information Management Group<br>
 * Date: 21-Jun-2008<br><br>
 */
public class OWL2Profile implements OWLProfile {


    private OWLObjectPropertyManager propertyManager;


    public String getName() {
        return "OWL 2";
    }


    public OWLProfileReport checkOntology(OWLOntology ontology, OWLOntologyManager manager) {
        propertyManager = new OWLObjectPropertyManager(manager, ontology);

        Set<ConstructNotAllowed> notAllowed = new HashSet<ConstructNotAllowed>();
        AxiomChecker checker = new AxiomChecker();
        for (OWLOntology ont : manager.getImportsClosure(ontology)) {
            for (OWLAxiom ax : ont.getAxioms()) {
                Set<ConstructNotAllowed> na = ax.accept(checker);
                notAllowed.addAll(na);
            }
        }

        // Property chain restrictions
        for(OWLObjectPropertyChainSubPropertyAxiom ax : ontology.getAxioms(AxiomType.PROPERTY_CHAIN_SUB_PROPERTY)) {
            if(ax.getPropertyChain().size() == 2) {
                // OPE1 = OPE2 = OPE
                // Can be transitive
                if(ax.isEncodingOfTransitiveProperty()) {
                    // What?
                }
            }
        }

        // No punning between data and object properties
        Set<URI> objectPropertyURIs = new HashSet<URI>();
        Set<URI> dataPropertyURIs = new HashSet<URI>();
        for (OWLOntology ont : manager.getImportsClosure(ontology)) {
            for (OWLObjectProperty prop : ont.getReferencedObjectProperties()) {
                objectPropertyURIs.add(prop.getURI());
            }
            for (OWLDataProperty prop : ont.getReferencedDataProperties()) {
                dataPropertyURIs.add(prop.getURI());
            }
        }
        for(URI uri : objectPropertyURIs) {
            if(dataPropertyURIs.contains(uri)) {
                notAllowed.add(new DataPropertyURIUsedAsObjectPropertyURI(uri));
            }
        }


        Set<OWLObjectPropertyExpression> simpleRoles = new HashSet<OWLObjectPropertyExpression>();


        Set<OWLObjectPropertyExpression> nonSimpleRoles = new HashSet<OWLObjectPropertyExpression>();

        for(OWLObjectProperty prop : ontology.getReferencedObjectProperties()) {
            if(!propertyManager.getNonSimpleProperties().contains(prop)) {
                simpleRoles.add(prop);
            }
            else {
                nonSimpleRoles.add(prop);
            }
        }
        return new OWL2ProfileReport(this, ontology.getURI(), notAllowed, nonSimpleRoles, simpleRoles);
    }

    private static Set<ConstructNotAllowed> getSet(ConstructNotAllowed na) {
        return Collections.singleton(na);
    }

    private static Set<ConstructNotAllowed> getSet() {
        return new HashSet<ConstructNotAllowed>(2);
    }

    private class AxiomChecker implements OWLObjectVisitorEx<Set<ConstructNotAllowed>> {


        public Set<ConstructNotAllowed> visit(OWLAntiSymmetricObjectPropertyAxiom axiom) {
            if(propertyManager.isNonSimple(axiom.getProperty())) {
                return getSet(new NonSimplePropertyInAntiSymmetricPropertyAxiom(axiom));
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLAxiomAnnotationAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLClassAssertionAxiom axiom) {
            Set<ConstructNotAllowed> na = axiom.getDescription().accept(this);
            if(na.isEmpty()) {
                return Collections.emptySet();
            }
            else {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
        }


        public Set<ConstructNotAllowed> visit(OWLDataPropertyAssertionAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataPropertyDomainAxiom axiom) {
            Set<ConstructNotAllowed> na = axiom.getDomain().accept(this);
            if (!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataPropertyRangeAxiom axiom) {
            Set<ConstructNotAllowed> na = axiom.getRange().accept(this);
            if (!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLDataSubPropertyAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDeclarationAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDifferentIndividualsAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDisjointClassesAxiom axiom) {
            Set<ConstructNotAllowed> na = getSet();
            for(OWLDescription desc : axiom.getDescriptions()) {
                Set<ConstructNotAllowed> descNa = desc.accept(this);
                na.addAll(descNa);
            }
            if(!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDisjointDataPropertiesAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDisjointObjectPropertiesAxiom axiom) {
            Set<OWLObjectPropertyExpression> nonSimpleProperties = new HashSet<OWLObjectPropertyExpression>();
            for(OWLObjectPropertyExpression prop : axiom.getProperties()) {
                if(propertyManager.isNonSimple(prop)) {
                    nonSimpleProperties.add(prop);
                }
            }
            if(!nonSimpleProperties.isEmpty()) {
                return getSet(new NonSimplePropertyInDisjointPropertiesAxiom(axiom, nonSimpleProperties));
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDisjointUnionAxiom axiom) {
            Set<ConstructNotAllowed> na = getSet();
            for(OWLDescription desc : axiom.getDescriptions()) {
                na.addAll(desc.accept(this));
            }
            if(!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLEntityAnnotationAxiom axiom) {
            Set<ConstructNotAllowed> na = axiom.getAnnotation().accept(this);
            if (!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLEquivalentClassesAxiom axiom) {
            Set<ConstructNotAllowed> na = getSet();
            for(OWLDescription desc : axiom.getDescriptions()) {
                na.addAll(desc.accept(this));
            }
            if(!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLEquivalentDataPropertiesAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLEquivalentObjectPropertiesAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLFunctionalDataPropertyAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLFunctionalObjectPropertyAxiom axiom) {
            if (propertyManager.isNonSimple(axiom.getProperty())) {
                return getSet(new NonSimplePropertyInFunctionalPropertyAxiom(axiom));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLImportsDeclaration axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLInverseFunctionalObjectPropertyAxiom axiom) {
            if(propertyManager.isNonSimple(axiom.getProperty())) {
                return getSet(new NonSimplePropertyInInverseFunctionalPropertyAxiom(axiom));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLInverseObjectPropertiesAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLIrreflexiveObjectPropertyAxiom axiom) {
            if(propertyManager.isNonSimple(axiom.getProperty())) {
                return getSet(new NonSimplePropertyInIrreflexivePropertyAxiom(axiom));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLNegativeDataPropertyAssertionAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLNegativeObjectPropertyAssertionAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectPropertyAssertionAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectPropertyChainSubPropertyAxiom axiom) {
            // Handled separately
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectPropertyDomainAxiom axiom) {
            Set<ConstructNotAllowed> na = axiom.getDomain().accept(this);
            if (!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectPropertyRangeAxiom axiom) {
            Set<ConstructNotAllowed> na = axiom.getRange().accept(this);
            if (!na.isEmpty()) {
                return getSet(new AxiomNotAllowed(na, axiom));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectSubPropertyAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLOntologyAnnotationAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLReflexiveObjectPropertyAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLSameIndividualsAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLSubClassAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLSymmetricObjectPropertyAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLTransitiveObjectPropertyAxiom axiom) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLRule rule) {
            for (SWRLAtom atom : rule.getBody()) {

            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLClass desc) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataAllRestriction desc) {
            Set<ConstructNotAllowed> na = desc.getFiller().accept(this);
            if (!na.isEmpty()) {
                return getSet(new DescriptionNotAllowed(na, desc));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLDataExactCardinalityRestriction desc) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataMaxCardinalityRestriction desc) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataMinCardinalityRestriction desc) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataSomeRestriction desc) {
            Set<ConstructNotAllowed> na = desc.getFiller().accept(this);
            if (!na.isEmpty()) {
                return getSet(new DescriptionNotAllowed(na, desc));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLDataValueRestriction desc) {
            Set<ConstructNotAllowed> na = desc.getValue().accept(this);
            if (!na.isEmpty()) {
                return getSet(new DescriptionNotAllowed(na, desc));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectAllRestriction desc) {
            Set<ConstructNotAllowed> na = desc.getFiller().accept(this);
            if (!na.isEmpty()) {
                return getSet(new DescriptionNotAllowed(na, desc));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectComplementOf desc) {
            Set<ConstructNotAllowed> na = desc.getOperand().accept(this);
            if (!na.isEmpty()) {
                return getSet(new DescriptionNotAllowed(na, desc));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectExactCardinalityRestriction desc) {
                if (propertyManager.isNonSimple(desc.getProperty())) {
                    return getSet(new NonSimplePropertyInCardinalityRestriction(desc));
                }
                else {
                    return Collections.emptySet();
                }

        }


        public Set<ConstructNotAllowed> visit(OWLObjectIntersectionOf desc) {
            for (OWLDescription op : desc.getOperands()) {
                Set<ConstructNotAllowed> na = op.accept(this);
                if (!na.isEmpty()) {
                    return getSet(new DescriptionNotAllowed(na, desc));
                }
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectMaxCardinalityRestriction desc) {
                if (propertyManager.isNonSimple(desc.getProperty())) {
                    return getSet(new NonSimplePropertyInCardinalityRestriction(desc));
                }
                else {
                    return Collections.emptySet();
                }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectMinCardinalityRestriction desc) {
                if (propertyManager.isNonSimple(desc.getProperty())) {
                    return getSet(new NonSimplePropertyInCardinalityRestriction(desc));
                }
                else {
                    return Collections.emptySet();
                }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectOneOf desc) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectSelfRestriction desc) {
            if(propertyManager.isNonSimple(desc.getProperty())) {
                return getSet(new NonSimplePropertyInExistsSelfRestriction(desc));
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectSomeRestriction desc) {
            Set<ConstructNotAllowed> na = desc.getFiller().accept(this);
            if (!na.isEmpty()) {
                return getSet(new DescriptionNotAllowed(na, desc));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLObjectUnionOf desc) {
            for (OWLDescription op : desc.getOperands()) {
                Set<ConstructNotAllowed> na = op.accept(this);
                if (!na.isEmpty()) {
                    return getSet(new DescriptionNotAllowed(na, desc));
                }
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectValueRestriction desc) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataComplementOf node) {
            Set<ConstructNotAllowed> na = node.getDataRange().accept(this);
            if (!na.isEmpty()) {
                return getSet(new DataRangeNotAllowed(na, node));
            }
            else {
                return Collections.emptySet();
            }
        }


        public Set<ConstructNotAllowed> visit(OWLDataOneOf node) {
            for (OWLConstant op : node.getValues()) {
                Set<ConstructNotAllowed> na = op.accept(this);
                if (!na.isEmpty()) {
                    return getSet(new DataRangeNotAllowed(na, node));
                }
            }
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataRangeFacetRestriction node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataRangeRestriction node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataType node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLTypedConstant node) {
            Set<ConstructNotAllowed> na = node.getDataType().accept(this);
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLUntypedConstant node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLDataProperty property) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectProperty property) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectPropertyInverse property) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLIndividual individual) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLConstantAnnotation annotation) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLObjectAnnotation annotation) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLAtomConstantObject node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLAtomDVariable node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLAtomIndividualObject node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLAtomIVariable node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLBuiltInAtom node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLClassAtom node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLDataRangeAtom node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLDataValuedPropertyAtom node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLDifferentFromAtom node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLObjectPropertyAtom node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(SWRLSameAsAtom node) {
            return Collections.emptySet();
        }


        public Set<ConstructNotAllowed> visit(OWLOntology ontology) {
            return Collections.emptySet();
        }
    }


    ///////////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    //  Not allowed constructs
    //
    ///////////////////////////////////////////////////////////////////////////////////////////////////////////


    public class AnnotationValueNotAllowed extends ConstructNotAllowed<OWLAnnotation> {


        public AnnotationValueNotAllowed(OWLAnnotation annotation) {
            super(annotation);
        }


        public AnnotationValueNotAllowed(ConstructNotAllowed cause, OWLAnnotation construct) {
            super(cause, construct);
        }
    }


    public class OWL2DataRangeNotAllowed extends DataRangeNotAllowed {

        public OWL2DataRangeNotAllowed(ConstructNotAllowed cause, OWLDataRange construct) {
            super(cause, construct);
        }


        public OWL2DataRangeNotAllowed(OWLDataRange construct) {
            super(construct);
        }
    }


    public class NonSimplePropertyInFunctionalPropertyAxiom extends AxiomNotAllowed<OWLFunctionalObjectPropertyAxiom> {

        public NonSimplePropertyInFunctionalPropertyAxiom(OWLFunctionalObjectPropertyAxiom construct) {
            super(construct);
        }
    }


    public class NonSimplePropertyInCardinalityRestriction extends DescriptionNotAllowed {

        public NonSimplePropertyInCardinalityRestriction(OWLObjectCardinalityRestriction construct) {
            super(construct);
        }
    }

    public class NonSimplePropertyInInverseFunctionalPropertyAxiom extends AxiomNotAllowed {

        public NonSimplePropertyInInverseFunctionalPropertyAxiom(OWLAxiom construct) {
            super(construct);
        }
    }

    public class NonSimplePropertyInAntiSymmetricPropertyAxiom extends AxiomNotAllowed {

        public NonSimplePropertyInAntiSymmetricPropertyAxiom(OWLAxiom construct) {
            super(construct);
        }
    }

    public class NonSimplePropertyInIrreflexivePropertyAxiom extends AxiomNotAllowed {

        public NonSimplePropertyInIrreflexivePropertyAxiom(OWLAxiom construct) {
            super(construct);
        }
    }


    public class NonSimplePropertyInDisjointPropertiesAxiom extends AxiomNotAllowed {

        private Set<OWLObjectPropertyExpression> props;

        public NonSimplePropertyInDisjointPropertiesAxiom(OWLAxiom construct, Set<OWLObjectPropertyExpression> properties) {
            super(construct);
            props = properties;
        }


        public Set<OWLObjectPropertyExpression> getProperties() {
            return props;
        }
    }

    public class NonSimplePropertyInExistsSelfRestriction extends  DescriptionNotAllowed {

        public NonSimplePropertyInExistsSelfRestriction(ConstructNotAllowed cause, OWLDescription construct) {
            super(cause, construct);
        }


        public NonSimplePropertyInExistsSelfRestriction(Set<ConstructNotAllowed> cause, OWLDescription construct) {
            super(cause, construct);
        }


        public NonSimplePropertyInExistsSelfRestriction(OWLDescription construct) {
            super(construct);
        }
    }

    public class DataPropertyURIUsedAsObjectPropertyURI extends ConstructNotAllowed<URI> {

        public DataPropertyURIUsedAsObjectPropertyURI(URI uri) {
            super(uri);
        }


        public String toString() {
            StringBuilder sb = new StringBuilder();
            sb.append("<");
            sb.append(getConstruct());
            sb.append(">  is used as both a data and object property URI");
            return sb.toString();
        }
    }
}
